/*
 * qscale: Quick, high-quality JPEG-to-JPEG scaler.
 * Copyright (C) 2008 Steinar H. Gunderson <sgunderson@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include "libqscale.h"

/* The number of pixels to process at a time when scaling vertically. */
#define CACHE_LINE_FACTOR 16

/* Whether to use SSE for horizontal scaling or not (requires SSE3). */
#define USE_HORIZONTAL_SSE 1

/* Whether to use SSE for vertical scaling or not (requires only SSE1). */
#define USE_VERTICAL_SSE 1

#if USE_VERTICAL_SSE
#undef CACHE_LINE_FACTOR
#define CACHE_LINE_FACTOR 16
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846264
#endif

qscale_img *qscale_load_jpeg(const char *filename)
{
	FILE *file = fopen(filename, "rb");
	qscale_img *img;
	if (file == NULL) {
		return NULL;
	}

	img = qscale_load_jpeg_from_stdio(file);

	fclose(file);
	return img;
}

qscale_img *qscale_load_jpeg_from_stdio(FILE *file)
{
	qscale_img *img = (qscale_img *)malloc(sizeof(qscale_img));
	if (img == NULL) {
		return NULL;
	}

	img->data_y = img->data_cb = img->data_cr = NULL;

	/* FIXME: Better error handling here (ie., return NULL). */
	struct jpeg_decompress_struct dinfo;
	struct jpeg_error_mgr jerr;
	dinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&dinfo);
	jpeg_stdio_src(&dinfo, file);
	jpeg_read_header(&dinfo, TRUE);
	dinfo.raw_data_out = TRUE;
	jpeg_start_decompress(&dinfo);
	
	/* We do not handle anything but YCbCr images (yet?). */
	if (dinfo.num_components != 3) {
		qscale_destroy(img);
		return NULL;
	}

	img->width = dinfo.image_width;
	img->height = dinfo.image_height;

	img->w0 = dinfo.image_width * dinfo.comp_info[0].h_samp_factor / dinfo.max_h_samp_factor;
	img->h0 = dinfo.image_height * dinfo.comp_info[0].v_samp_factor / dinfo.max_v_samp_factor;

	img->w1 = dinfo.image_width * dinfo.comp_info[1].h_samp_factor / dinfo.max_h_samp_factor;
	img->h1 = dinfo.image_height * dinfo.comp_info[1].v_samp_factor / dinfo.max_v_samp_factor;

	img->w2 = dinfo.image_width * dinfo.comp_info[2].h_samp_factor / dinfo.max_h_samp_factor;
	img->h2 = dinfo.image_height * dinfo.comp_info[2].v_samp_factor / dinfo.max_v_samp_factor;

	img->samp_h0 = dinfo.comp_info[0].h_samp_factor;
	img->samp_v0 = dinfo.comp_info[0].v_samp_factor;

	img->samp_h1 = dinfo.comp_info[1].h_samp_factor;
	img->samp_v1 = dinfo.comp_info[1].v_samp_factor;

	img->samp_h2 = dinfo.comp_info[2].h_samp_factor;
	img->samp_v2 = dinfo.comp_info[2].v_samp_factor;

	img->data_y  = (JSAMPLE*)memalign(16, dinfo.comp_info[0].height_in_blocks * dinfo.comp_info[0].width_in_blocks * DCTSIZE * DCTSIZE);
	img->data_cb = (JSAMPLE*)memalign(16, dinfo.comp_info[1].height_in_blocks * dinfo.comp_info[1].width_in_blocks * DCTSIZE * DCTSIZE);
	img->data_cr = (JSAMPLE*)memalign(16, dinfo.comp_info[2].height_in_blocks * dinfo.comp_info[2].width_in_blocks * DCTSIZE * DCTSIZE);

	if (img->data_y == NULL || img->data_cb == NULL || img->data_cr == NULL) {
		qscale_destroy(img);
		return NULL;
	}

	int total_lines = 0, blocks = 0;
        while (total_lines < dinfo.comp_info[0].height_in_blocks * DCTSIZE) {
                unsigned max_lines = dinfo.max_v_samp_factor * DCTSIZE;

                JSAMPROW y_row_ptrs[max_lines];
                JSAMPROW cb_row_ptrs[max_lines];
                JSAMPROW cr_row_ptrs[max_lines];
                JSAMPROW* ptrs[] = { y_row_ptrs, cb_row_ptrs, cr_row_ptrs };

		int i;
                for (i = 0; i < max_lines; ++i) {
                        y_row_ptrs[i]  = img->data_y  + (i+blocks*DCTSIZE*dinfo.comp_info[0].v_samp_factor) * dinfo.comp_info[0].width_in_blocks * DCTSIZE;
                        cb_row_ptrs[i] = img->data_cb + (i+blocks*DCTSIZE*dinfo.comp_info[1].v_samp_factor) * dinfo.comp_info[1].width_in_blocks * DCTSIZE;
                        cr_row_ptrs[i] = img->data_cr + (i+blocks*DCTSIZE*dinfo.comp_info[2].v_samp_factor) * dinfo.comp_info[2].width_in_blocks * DCTSIZE;
                }

                total_lines += max_lines;
                ++blocks;

                if (jpeg_read_raw_data(&dinfo, ptrs, max_lines) == 0)
                        break;
        }

	jpeg_destroy_decompress(&dinfo);
	return img;
}

void qscale_destroy(qscale_img *img)
{
	free(img->data_y);
	free(img->data_cb);
	free(img->data_cr);
	free(img);
}


static double sinc(double x)
{
	static const double cutoff = 1.220703668e-4;  /* sqrt(sqrt(eps)) */

	if (abs(x) < cutoff) {
		/* For small |x|, use Taylor series instead */
		const double x2 = x * x;
		const double x4 = x2 * x2;

		return 1.0 - x2 / 6.0 + x4 / 120.0;
	} else {
		return sin(x) / x;
	}
}

static double lanczos_tap(double x)
{
	if (x < -3.0 || x > 3.0)
		return 0.0;
	if (x < 0.0)
		return sinc(-x*M_PI) * sinc(-x*M_PI / 3.0);
	else
		return sinc(x*M_PI) * sinc(x*M_PI / 3.0);
}

struct pix_desc {
	unsigned start, end;
	unsigned startcoeff;
};

static void hscale(float *pix, unsigned char *npix, unsigned w, unsigned h, unsigned nw, unsigned sstride, unsigned dstride)
{
	struct pix_desc *pd = (struct pix_desc *)malloc(nw * sizeof(struct pix_desc));
	int size_coeffs = 8;
	float *coeffs = (float *)malloc(size_coeffs * sizeof(float));
	int num_coeffs = 0;
	int x, y;
	double sf = (double)w / (double)nw;
	double support = (w > nw) ? (3.0 * sf) : (3.0 / sf);

	/* calculate the filter */
	for (x = 0; x < nw; ++x) {
		int start = ceil(x * sf - support);
		int end = floor(x * sf + support);
		int sx;
		double sum = 0.0;

		if (start < 0) {
			start = 0;
		}
		if (end > w - 1) {
			end = w - 1;
		}

#if USE_HORIZONTAL_SSE
		/* round up so we get a multiple of four for the SSE code */
		int num = (end - start + 1);
		if (num % 4 != 0) {
			/* prefer aligning it if possible */
			if (start % 4 != 0 && start % 4 <= num % 4) {
				num += start % 4;
				start -= start % 4;
			}
			if (num % 4 != 0) {
				end += 4 - (num % 4);
			}
		}
#endif

		pd[x].start = start;
		pd[x].end = end;
		pd[x].startcoeff = num_coeffs;

		for (sx = start; sx <= end; ++sx) {
			double nd = (w > nw) ? (sx/sf - x) : (sx - x*sf);
			double f = lanczos_tap(nd);
			if (num_coeffs == size_coeffs) {
				size_coeffs <<= 1;
				coeffs = (float *)realloc(coeffs, size_coeffs * sizeof(float));
			}

			coeffs[num_coeffs++] = f;
			sum += f;
		}

		for (sx = start; sx <= end; ++sx) {
			coeffs[pd[x].startcoeff + sx - start] /= sum;
		}
	}

	for (y = 0; y < h; ++y) {
		float *sptr = pix + y*sstride;
		unsigned char *dptr = npix + y*dstride;
		unsigned char ch;
		for (x = 0; x < nw; ++x) {
#if USE_HORIZONTAL_SSE
			int result;
			float acc;
			long tmp;
			static const float low = 0.0, high = 255.0;
			__asm__ (
				"pxor %1, %1               \n"
				"xor %2, %2                \n"
				"0:                        \n"
				"movups (%4,%2),%%xmm1     \n"
				"movups (%3,%2),%%xmm2     \n"
				"mulps %%xmm2,%%xmm1       \n"
				"addps %%xmm1,%1           \n"
				"add $16,%2                \n"
				"dec %5                    \n"
				"jnz 0b                    \n"
				"haddps %1,%1              \n"
				"haddps %1,%1              \n"
				"maxss %6,%1               \n"
				"minss %7,%1               \n"
				"cvtss2si %1,%0            \n"
				: "=r" (result),
				  "=&x" (acc),
				  "=&r" (tmp)
				: "r" (&coeffs[pd[x].startcoeff]),
				  "r" (&sptr[pd[x].start]),
				  "r" ((pd[x].end - pd[x].start + 1)/4),
				  "m" (low),
				  "m" (high)
				: "memory", "xmm1", "xmm2"
			);

			*dptr++ = (unsigned char)result;
#else
			float acc = 0.0;
			float *cf = &coeffs[pd[x].startcoeff];
			unsigned sx;
			
			for (sx = pd[x].start; sx <= pd[x].end; ++sx) {
				acc += sptr[sx] * *cf++;
			}

			if (acc < 0.0)
				ch = 0;
			else if (acc > 255.0)
				ch = 255;
			else
				ch = (unsigned char)acc;
			*dptr++ = ch;
#endif
		}
		ch = dptr[-1];
		for ( ; x < dstride; ++x) {
			*dptr++ = ch;
		}
	}
}

static void vscale(unsigned char *pix, float *npix, unsigned w, unsigned h, unsigned nh, unsigned dstride)
{
	struct pix_desc *pd = (struct pix_desc *)malloc(nh * sizeof(struct pix_desc));
	int size_coeffs = 8;
	float *coeffs = (float *)malloc(size_coeffs * sizeof(float));
	int num_coeffs = 0;
	int x, y, sy;
	double sf = (double)h / (double)nh;
	double support = (h > nh) ? (3.0 * sf) : (3.0 / sf);

	/* calculate the filter */
	for (y = 0; y < nh; ++y) {
		int start = ceil(y * sf - support);
		int end = floor(y * sf + support);
		double sum = 0.0;

		if (start < 0) {
			start = 0;
		}
		if (end > h - 1) {
			end = h - 1;
		}

		pd[y].start = start;
		pd[y].end = end;
		pd[y].startcoeff = num_coeffs;

		for (sy = start; sy <= end; ++sy) {
			double nd = (h > nh) ? (sy/sf - y) : (sy - y*sf);
			double f = lanczos_tap(nd);
			if (num_coeffs == size_coeffs) {
				size_coeffs <<= 1;
				coeffs = (float *)realloc(coeffs, size_coeffs * sizeof(float));
			}
			
			coeffs[num_coeffs++] = f;
			sum += f;
		}

		for (sy = start; sy <= end; ++sy) {
			coeffs[pd[y].startcoeff + sy - start] /= sum;
		}
	}

#if CACHE_LINE_FACTOR > 1
	for (x = 0; x < (w/CACHE_LINE_FACTOR) * CACHE_LINE_FACTOR; x += CACHE_LINE_FACTOR) {
		unsigned char *sptr = pix + x;
		float *dptr = npix + x;
		for (y = 0; y < nh; ++y) {
#if USE_VERTICAL_SSE
			/*
			 * xmm0 - xmm3: acc[0..15]
			 * xmm4: current filter coefficient
			 * xmm5, xmm6, xmm7: scratchpad
			 */
			__asm__ (
				/* clear */
				"pxor %%xmm0, %%xmm0          \n"
				"pxor %%xmm1, %%xmm1          \n"
				"pxor %%xmm2, %%xmm2          \n"
				"pxor %%xmm3, %%xmm3          \n"

				/* main loop */
				"0:                           \n"
				
				/* a zero is useful during unpacking */
				"pxor %%xmm4, %%xmm4          \n"
				
				/* fetch all 16 source bytes */
				"movups (%0), %%xmm5          \n"
				"prefetcht0 (%0,%3,4)         \n"

				/* unpack into words (xmm5, xmm7) */
				"movaps %%xmm5, %%xmm7        \n"
				"punpcklbw %%xmm4, %%xmm5     \n"
				"punpckhbw %%xmm4, %%xmm7     \n"

				/* unpack xmm5 into dwords (xmm5, xmm6) */
				"movaps %%xmm5, %%xmm6        \n"
				"punpcklwd %%xmm4, %%xmm5     \n"
				"punpckhwd %%xmm4, %%xmm6     \n"

				/* convert xmm5, xmm6 to floats */
				"cvtdq2ps %%xmm5, %%xmm5      \n"
				"cvtdq2ps %%xmm6, %%xmm6      \n"

				/* fetch the coefficient */
				"movss (%2), %%xmm4           \n"
				"shufps $0x0, %%xmm4, %%xmm4  \n"

				/* do the muls for xmm5 and xmm6 */
				"mulps %%xmm4, %%xmm5         \n"
				"mulps %%xmm4, %%xmm6         \n"
				"addps %%xmm5, %%xmm0         \n"
				"addps %%xmm6, %%xmm1         \n"

				/* get the zero back again */
				"pxor %%xmm4, %%xmm4          \n"

				/* unpack xmm7 into dwords (xmm7, xmm6) */
				"movaps %%xmm7, %%xmm6        \n"
				"punpcklwd %%xmm4, %%xmm7     \n"
				"punpckhwd %%xmm4, %%xmm6     \n"

				/* convert xmm7, xmm6 to floats */
				"cvtdq2ps %%xmm7, %%xmm7      \n"
				"cvtdq2ps %%xmm6, %%xmm6      \n"

				/* fetch the coefficient */
				"movss (%2), %%xmm4           \n"
				"shufps $0x0, %%xmm4, %%xmm4  \n"

				/* do the second set of muls */
				"mulps %%xmm4, %%xmm7         \n"
				"mulps %%xmm4, %%xmm6         \n"
				"addps %%xmm7, %%xmm2         \n"
				"addps %%xmm6, %%xmm3         \n"

				/* move along, and loop */
				"add $4, %2                   \n"
				"add %3, %0                   \n"
				"dec %1                       \n"
				"jnz 0b                       \n"

				/* store the values */
				"movaps %%xmm0, (%4)          \n"
				"movaps %%xmm1, 16(%4)        \n"
				"movaps %%xmm2, 32(%4)        \n"
				"movaps %%xmm3, 48(%4)        \n"
				: :
				"r" (&sptr[pd[y].start * w]),        /* 0: srcptr base */
				"r" (pd[y].end - pd[y].start + 1),   /* 1: filter len */
				"r" (&coeffs[pd[y].startcoeff]),     /* 2: coeffs base */
				"r" ((long)w),                       /* 3: stride */
				"r" (dptr)                           /* 4: dstptr base */
				: "memory", "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7"
			);
#else
			int i;
			float acc[CACHE_LINE_FACTOR];
			for (i = 0; i < CACHE_LINE_FACTOR; ++i)
				acc[i] = 0.0;
			float *cf = &coeffs[pd[y].startcoeff];
			unsigned sy;
		
			for (sy = pd[y].start; sy <= pd[y].end; ++sy) {
				for (i = 0; i < CACHE_LINE_FACTOR; ++i) {
					acc[i] += sptr[sy * w + i] * *cf;
				}
				++cf;
			}

			for (i = 0; i < CACHE_LINE_FACTOR; ++i) {
				dptr[i] = acc[i];
			}
#endif
			dptr += dstride;
		}
	}
	for (x = (x/CACHE_LINE_FACTOR)*CACHE_LINE_FACTOR; x < w; ++x) {
#else
	for (x = 0; x < w; ++x) {
#endif
		unsigned char *sptr = pix + x;
		float *dptr = npix + x;
		for (y = 0; y < nh; ++y) {
			float acc = 0.0;
			float *cf = &coeffs[pd[y].startcoeff];
			unsigned sy;
			
			for (sy = pd[y].start; sy <= pd[y].end; ++sy) {
				acc += sptr[sy * w] * *cf++;
			}

			*dptr = acc;
			dptr += dstride;
		}
	}
}

qscale_img *qscale_clone(const qscale_img *img)
{
	qscale_img *dst = (qscale_img *)malloc(sizeof(qscale_img));
	if (dst == NULL) {
		return NULL;
	}

	*dst = *img;

	unsigned dstride0 = (dst->w0 + DCTSIZE-1) & ~(DCTSIZE-1);
	unsigned dstride1 = (dst->w1 + DCTSIZE-1) & ~(DCTSIZE-1);
	unsigned dstride2 = (dst->w2 + DCTSIZE-1) & ~(DCTSIZE-1);

	/* FIXME: handle out-of-memory gracefully */
	{
		dst->data_y = (unsigned char *)malloc(dst->h0 * dstride0);
		memcpy(dst->data_y, img->data_y, dst->h0 * dstride0);
	}
	{
		dst->data_cb = (unsigned char *)malloc(dst->h1 * dstride1);
		memcpy(dst->data_cb, img->data_cb, dst->h0 * dstride0);
	}
	{
		dst->data_cr = (unsigned char *)malloc(dst->h1 * dstride1);
		memcpy(dst->data_cr, img->data_cr, dst->h0 * dstride0);
	}

	return dst;
}

qscale_img *qscale_scale(qscale_img *src, unsigned width, unsigned height, unsigned samp_h0, unsigned samp_v0, unsigned samp_h1, unsigned samp_v1, unsigned samp_h2, unsigned samp_v2, enum qscale_scaling_filter scaling_filter)
{
	qscale_img *dst = (qscale_img *)malloc(sizeof(qscale_img));
	if (dst == NULL) {
		return NULL;
	}

	dst->width = width;
	dst->height = height;

	unsigned max_samp_h, max_samp_v;
        max_samp_h = samp_h0;
        if (samp_h1 > max_samp_h)
                max_samp_h = samp_h1;
        if (samp_h2 > max_samp_h)
                max_samp_h = samp_h2;

        max_samp_v = samp_v0;
        if (samp_v1 > max_samp_v)
                max_samp_v = samp_v1;
        if (samp_v2 > max_samp_v)
                max_samp_v = samp_v2;

	dst->w0 = width * samp_h0 / max_samp_h;
	dst->h0 = height * samp_v0 / max_samp_v;

	dst->w1 = width * samp_h1 / max_samp_h;
	dst->h1 = height * samp_v1 / max_samp_v;

	dst->w2 = width * samp_h2 / max_samp_h;
	dst->h2 = height * samp_v2 / max_samp_v;

	dst->samp_h0 = samp_h0;
	dst->samp_v0 = samp_v0;

	dst->samp_h1 = samp_h1;
	dst->samp_v1 = samp_v1;

	dst->samp_h2 = samp_h2;
	dst->samp_v2 = samp_v2;

	unsigned dstride0 = (dst->w0 + DCTSIZE-1) & ~(DCTSIZE-1);
	unsigned dstride1 = (dst->w1 + DCTSIZE-1) & ~(DCTSIZE-1);
	unsigned dstride2 = (dst->w2 + DCTSIZE-1) & ~(DCTSIZE-1);

	unsigned sstride0 = (src->w0 + DCTSIZE-1) & ~(DCTSIZE-1);
	unsigned sstride1 = (src->w1 + DCTSIZE-1) & ~(DCTSIZE-1);
	unsigned sstride2 = (src->w2 + DCTSIZE-1) & ~(DCTSIZE-1);

	/* FIXME: handle out-of-memory gracefully */
	{
		float *npix = (float*)memalign(16, sstride0 * dst->h0 * sizeof(float));
		vscale(src->data_y, npix, sstride0, src->h0, dst->h0, sstride0);
		dst->data_y = (unsigned char *)malloc(dst->h0 * dstride0);
		hscale(npix, dst->data_y, src->w0, dst->h0, dst->w0, sstride0, dstride0);
		free(npix);
	}
	{
		float *npix = (float*)memalign(16, sstride1 * dst->h1 * sizeof(float));
		vscale(src->data_cr, npix, sstride1, src->h1, dst->h1, sstride1);
		dst->data_cr = (unsigned char *)malloc(dst->h1 * dstride1);
		hscale(npix, dst->data_cr, src->w1, dst->h1, dst->w1, sstride1, dstride1);
		free(npix);
	}
	{
		float *npix = (float*)memalign(16, sstride2 * dst->h2 * sizeof(float));
		vscale(src->data_cb, npix, sstride2, src->h2, dst->h2, sstride2);
		dst->data_cb = (unsigned char *)malloc(dst->h2 * dstride2);
		hscale(npix, dst->data_cb, src->w2, dst->h2, dst->w2, sstride2, dstride2);
		free(npix);
	}

	return dst;
}

int qscale_save_jpeg(const qscale_img *img, const char *filename, unsigned jpeg_quality, enum qscale_jpeg_mode jpeg_mode)
{
	FILE *file = fopen(filename, "wb");
	if (file == NULL) {
		return -1;
	}

	int err = qscale_save_jpeg_to_stdio(img, file, jpeg_quality, jpeg_mode);

	fclose(file);
	return err;
}

int qscale_save_jpeg_to_stdio(const qscale_img *img, FILE *file, unsigned jpeg_quality, enum qscale_jpeg_mode jpeg_mode)
{
        struct jpeg_compress_struct cinfo;
        struct jpeg_error_mgr jerr;
        cinfo.err = jpeg_std_error(&jerr);
        jpeg_create_compress(&cinfo);
        jpeg_stdio_dest(&cinfo, file);
        cinfo.input_components = 3;
        jpeg_set_defaults(&cinfo);
        jpeg_set_quality(&cinfo, jpeg_quality, FALSE);
        cinfo.image_width = img->width;
        cinfo.image_height = img->height;
        cinfo.raw_data_in = TRUE;
        jpeg_set_colorspace(&cinfo, JCS_YCbCr);
        cinfo.comp_info[0].h_samp_factor = img->samp_h0;
        cinfo.comp_info[0].v_samp_factor = img->samp_v0;
        cinfo.comp_info[1].h_samp_factor = img->samp_h1;
        cinfo.comp_info[1].v_samp_factor = img->samp_v1;
        cinfo.comp_info[2].h_samp_factor = img->samp_h2;
        cinfo.comp_info[2].v_samp_factor = img->samp_v2;
        jpeg_start_compress(&cinfo, TRUE);

        unsigned dstride0 = (img->w0 + DCTSIZE-1) & ~(DCTSIZE-1);
        unsigned dstride1 = (img->w1 + DCTSIZE-1) & ~(DCTSIZE-1);
        unsigned dstride2 = (img->w2 + DCTSIZE-1) & ~(DCTSIZE-1);

        int total_lines = 0;
        int blocks = 0;
        while (total_lines < cinfo.comp_info[0].height_in_blocks * DCTSIZE) {
                unsigned max_lines = cinfo.max_v_samp_factor * DCTSIZE;

                JSAMPROW y_row_ptrs[max_lines];
                JSAMPROW cb_row_ptrs[max_lines];
                JSAMPROW cr_row_ptrs[max_lines];
                JSAMPROW* ptrs[] = { y_row_ptrs, cb_row_ptrs, cr_row_ptrs };
                int i;

                for (i = 0; i < max_lines; ++i) {
                        /* simple edge extension */
                        int yline = i + blocks*DCTSIZE*cinfo.comp_info[0].v_samp_factor;
                        if (yline > img->h0 - 1)
                                yline = img->h0 - 1;

                        int cbline = i + blocks*DCTSIZE*cinfo.comp_info[1].v_samp_factor;
			if (cbline > img->h1 - 1)
                                cbline = img->h1 - 1;

                        int crline = i + blocks*DCTSIZE*cinfo.comp_info[2].v_samp_factor;
                        if (crline > img->h2 - 1)
                                crline = img->h2 - 1;

                        y_row_ptrs[i]  = img->data_y  + yline * dstride0;
                        cb_row_ptrs[i] = img->data_cb + cbline * dstride1;
                        cr_row_ptrs[i] = img->data_cr + crline * dstride2;
                }

                total_lines += max_lines;
                ++blocks;

                jpeg_write_raw_data(&cinfo, ptrs, max_lines);
        }
        jpeg_finish_compress(&cinfo);
        jpeg_destroy_compress(&cinfo);

	return 0;
}
